% This is the main run file for estimating the marginal likelihood of the
% hybrid time-varying VARs in Chan and Eisenstat (2018). 
%
% This code is free to use for academic purposes only, provided that the 
% paper is cited as:
%
% Chan, J.C.C. and Eisenstat, E. (2018). Comparing Hybrid Time-Varying 
% Parameter VARs, Economics Letters, 171, 1-5.
%
% This code comes without technical support of any kind.  It is expected to
% reproduce the results reported in the paper. Under no circumstances will
% the authors be held responsible for any use (or misuse) of this code in
% any way.

clear; clc;
warning off;

nsim = 20000;
burnin = 5000;
M = 20000;   % number of replications in ML estimation
p = 2;       % number of lags 
is_tv = [0 1 0]'; % n-vector that denotes which equations have TVP; must have the same dim as the VAR
                  % e.g., [1 0 1]: eq 1 and 3 have TVP, but eq 2 has const parameters
is_SR = 0;   % 1: replace Fed Funds rate with Wu-Xi shadow rate; 0: otherwise
is_SS = 0;   % 1: use sample from 1969Q3 - 1999Q4; 0: full sample

% load US data
USdata = xlsread('USdata_2017Q4.xlsx','B28:E281'); % 1954Q3-2017Q4
data = USdata(:,[1 3 4]); % GDP deflator growth, real GDP growth, Fed funds rate
if is_SR
    Wu_Xi = xlsread('Wu-Xia-shadow-rates.xls','B1:B672'); % 1960M1-2015M12
    S_rate = mean(reshape(Wu_Xi,3,length(Wu_Xi)/3))';     % 1960Q1-2015Q4
    data(23:246,3) = S_rate;  % replace the Fed funds rate by Wu-Xi shadow rate    
end
Y0 = data(1:4,:); shortY = data(5:end,:); % full sample
if is_SS 
    Y0 = data(61:64,:); shortY = data(65:182,:); % 1969Q3 - 1999Q4    
end
[T,n] = size(shortY);
Y = reshape(shortY',T*n,1);
k_alp = n*(n-1)/2;       % dimension of the impact matrix
k_beta = n^2*p + n;      % number of VAR coefficients    
n_tv = sum(is_tv);       % # of time-varying equations

%% prior
abeta0 = zeros(k_beta,1); Vbeta0 = 10*ones(k_beta,1);
aalp0 = zeros(k_alp,1); Valp0 = 10*ones(k_alp,1);
ah0 = zeros(n,1); Vh0 = 10*ones(n,1);
nuh0 = 5*ones(n,1); Sh0 = .01*ones(n,1).*(nuh0-1);
nubeta0 = 5*ones(k_beta,1); Sbeta0 = .01^2*ones(k_beta,1).*(nubeta0-1); 
Sbeta0(1:n*p+1:end) = .1^2*(nubeta0(1:n*p+1:end)-1); % intercepts
Sbeta0 = Sbeta0.*kron(is_tv,ones(n*p+1,1)); % set to 0 for time-invariant equations
nualp0 = 5*ones(k_alp,1); Salp0 = .01^2*ones(k_alp,1).*(nualp0-1);
count = 0;
for ii=2:n
    if is_tv(ii) == 0
        Salp0(count+1:count+ii-1) = 0; % set to 0 for time-invariant equations
    end
    count = count + ii - 1;
end 

%% compute and define a few things
tmpY = [Y0(end-p+1:end,:); shortY];
X2 = zeros(T,n*p); 
for ii=1:p
    X2(:,(ii-1)*n+1:ii*n) = tmpY(p-ii+1:end-ii,:);
end
X2 = [ones(T,1) X2];
X1 = zeros(n*T,k_alp);
count = 0;
for ii=2:n
    X1(ii:n:end,count+1:count+ii-1) = -shortY(:,1:ii-1);
    count = count + ii-1;
end

idx_b_tv = (kron(is_tv,ones(n*p+1,1))==1); % index for time-varying betas
idx_a_tv = false(k_alp,1); % construct index for time-varying alphas
count = 0;
for j=2:n
    if is_tv(j) == 1
        idx_a_tv( count+1:count+j-1) = true;
    end
    count = count + j-1;
end

cpri = -.5*(n+k_beta+k_alp)*log(2*pi) -.5*sum(log(Vbeta0)) -.5*sum(log(Vh0)) ...
    + nubeta0(idx_b_tv)'*log(Sbeta0(idx_b_tv)) - sum(gammaln(nubeta0(idx_b_tv))) ...
    + nualp0(idx_a_tv)'*log(Salp0(idx_a_tv)) - sum(gammaln(nualp0(idx_a_tv))) ...
    + nuh0'*log(Sh0) - sum(gammaln(nuh0));
prior = @(sb,sa,sh,b0,a0,c0) cpri -.5*(b0-abeta0)'*((b0-abeta0)./Vbeta0) ...
    -.5*(a0-aalp0)'*((a0-aalp0)./Valp0) -.5*((c0-ah0)./Vh0)'*(c0-ah0) ...
    -(nubeta0(idx_b_tv)+1)'*log(sb(idx_b_tv)) -sum(Sbeta0(idx_b_tv)./sb(idx_b_tv)) ...
    -(nualp0(idx_a_tv)+1)'*log(sa(idx_a_tv)) -sum(Salp0(idx_a_tv)./sa(idx_a_tv)) ...
    -(nuh0+1)'*log(sh) -sum(Sh0./sh);


%% initialize for storage
store_alp = zeros(nsim,T,k_alp);
store_beta = zeros(nsim,T,k_beta);
store_h = zeros(nsim,T,n);
store_Sigbeta = zeros(nsim,k_beta);
store_Sigalp = zeros(nsim,k_alp);
store_Sigh = zeros(nsim,n);
store_beta0 = zeros(nsim,k_beta);
store_alp0 = zeros(nsim,k_alp);
store_h0 = zeros(nsim,n);

%% initialize the Markov chain
Sigbeta = zeros(k_beta,1);
Sigbeta(idx_b_tv) = .01*rand(k_beta/n*n_tv,1);
Sigalp = zeros(k_alp,1);
Sigalp(idx_a_tv) = .01*rand(sum(idx_a_tv),1);
Sigh = .1*rand(n,1);
beta0 = abeta0;
alp0 = aalp0;
h0 = log(var(shortY))';
h = repmat(h0',T,1);

%% MCMC starts here
randn('seed',sum(clock*100)); rand('seed',sum(clock*1000));
disp(['Starting MCMC for hybrid TVP-VAR-SV with ID: ' num2str(is_tv') '.... ']);
start_time = clock;
 
for isim = 1:nsim + burnin  
        % sample alp and beta - equation by equation    
    count = 0;
    U = zeros(T,n);
    for ii=1:n 
        ki = n*p+1+ii-1;
        if is_tv(ii) == 1
            bigXi = SURform([X2 -shortY(:,1:ii-1)]);            
            Hthetai = speye(T*ki) - sparse(ki+1:T*ki,1:(T-1)*ki,ones((T-1)*ki,1),T*ki,T*ki); 
            Sigthetai = [Sigbeta((ii-1)*k_beta/n+1:ii*k_beta/n); Sigalp(count+1:count+ii-1)];        
            thetai0 = [beta0((ii-1)*k_beta/n+1:ii*k_beta/n); alp0(count+1:count+ii-1)];                
            XiSig = bigXi'*sparse(1:T,1:T,exp(-h(:,ii)));
            HiSH_thetai = Hthetai'*sparse(1:T*ki,1:T*ki,repmat(1./Sigthetai,T,1))*Hthetai;
            Kthetai = HiSH_thetai + XiSig*bigXi;
            thetai_hat = Kthetai\(HiSH_thetai*kron(ones(T,1),thetai0) + XiSig*shortY(:,ii));        
            thetai = thetai_hat + chol(Kthetai,'lower')'\randn(T*ki,1);
            Thetai = reshape(thetai,ki,T)';
            beta(:,(ii-1)*k_beta/n+1:ii*k_beta/n) = Thetai(:,1:k_beta/n);
            alp(:,count+1:count+ii-1) = Thetai(:,k_beta/n+1:end);
        else
            bigXi = [X2 -shortY(:,1:ii-1)];
            XiSig = bigXi'*sparse(1:T,1:T,exp(-h(:,ii)));
            Vthetai = [Vbeta0((ii-1)*k_beta/n+1:ii*k_beta/n); Valp0(count+1:count+ii-1)];
            thetai0 = [beta0((ii-1)*k_beta/n+1:ii*k_beta/n); alp0(count+1:count+ii-1)];                
            Kthetai = Vthetai + XiSig*bigXi;
            thetai_hat = Kthetai\(thetai0./Vthetai + XiSig*shortY(:,ii));
            thetai = thetai_hat + chol(Kthetai,'lower')'\randn(ki,1);
            betai = thetai(1:k_beta/n); alpi = thetai(k_beta/n+1:end);
            beta0((ii-1)*k_beta/n+1:ii*k_beta/n) = betai;
            alp0(count+1:count+ii-1) = alpi;
            beta(:,(ii-1)*k_beta/n+1:ii*k_beta/n) = repmat(betai',T,1);
            alp(:,count+1:count+ii-1) = repmat(alpi',T,1);
        end
        count = count + ii-1;
        U(:,ii) = shortY(:,ii) - bigXi*thetai;
    end 
    
        % sample h    
    for ii = 1:n
        Ystar = log(U(:,ii).^2 + .0001);
        h(:,ii) = SVRW(Ystar,h(:,ii),Sigh(ii),h0(ii));
    end 

        % sample beta0    
    Kbeta0_tv = sparse(1:k_beta/n*n_tv,1:k_beta/n*n_tv,1./Sigbeta(idx_b_tv)+ 1./Vbeta0(idx_b_tv));
    beta0_tv_hat = Kbeta0_tv\(abeta0(idx_b_tv)./Vbeta0(idx_b_tv) + beta(1,idx_b_tv)'./Sigbeta(idx_b_tv));
    beta0(idx_b_tv) = beta0_tv_hat + chol(Kbeta0_tv,'lower')'\randn(k_beta/n*n_tv,1);
    
        % sample alp0
    Kalp0_tv = sparse(1:sum(idx_a_tv),1:sum(idx_a_tv),1./Sigalp(idx_a_tv) + 1./Valp0(idx_a_tv));
    alp0_tv_hat = Kalp0_tv\(aalp0(idx_a_tv)./Valp0(idx_a_tv) + alp(1,idx_a_tv)'./Sigalp(idx_a_tv));
    alp0(idx_a_tv) = alp0_tv_hat + chol(Kalp0_tv,'lower')'\randn(sum(idx_a_tv),1);    

        % sample h0
    Kh0 = sparse(1:n,1:n,1./Sigh + 1./Vh0);
    h0_hat = Kh0\(ah0./Vh0 + h(1,:)'./Sigh);
    h0 = h0_hat + chol(Kh0,'lower')'\randn(n,1);
    
        % sample Sigbeta
    E_beta = beta(:,idx_b_tv) - [beta0(idx_b_tv)';beta(1:T-1,idx_b_tv)];
    Sigbeta(idx_b_tv) = 1./gamrnd(nubeta0(idx_b_tv)+T/2, 1./(Sbeta0(idx_b_tv) + sum(E_beta.^2)'/2));
    
        % sample Sigalp
    E_alp = alp(:,idx_a_tv) - [alp0(idx_a_tv)';alp(1:T-1,idx_a_tv)];
    Sigalp(idx_a_tv) = 1./gamrnd(nualp0(idx_a_tv)+T/2, 1./(Salp0(idx_a_tv) + sum(E_alp.^2)'/2));
      
        % sample Sigh
    E_h = h - [h0';h(1:T-1,:)];
    Sigh = 1./gamrnd(nuh0+T/2, 1./(Sh0 + sum(E_h.^2)'/2));
    
    if isim > burnin
        isave = isim - burnin;
        store_h(isave,:,:) = h; 
        store_alp(isave,:,:) = alp;
        store_beta(isave,:,:) = beta;
        store_Sigbeta(isave,:) = Sigbeta';
        store_Sigalp(isave,:) = Sigalp';
        store_Sigh(isave,:) = Sigh';   
        store_beta0(isave,:) = beta0';
        store_alp0(isave,:) = alp0';
        store_h0(isave,:) = h0';        
    end
    
    if (mod(isim, 5000) == 0)
        disp([num2str(isim) ' loops... ']);
    end 
    
end

disp( ['MCMC takes '  num2str( etime( clock, start_time) ) ' seconds' ] );
disp(' ' );

alp_hat = squeeze(mean(store_alp));
alp_CI = quantile(store_alp,[.05 .95]);
beta_hat = squeeze(mean(store_beta));
beta_CI = quantile(store_beta,[.05 .95]);
h_hat = squeeze(mean(exp(store_h/2)));
h_CI = quantile(exp(store_h/2),[.05 .95]);

[ml, mlstd] = ml_hybrid_tvpsv(shortY,store_Sigbeta,store_Sigalp,store_Sigh,...
    store_beta0,store_alp0,store_h0,prior,X2,is_tv,M);
 
disp(' ')
fprintf('log marginal likelihood: %.1f (%.2f)\n', ml, mlstd);
disp(' ' );
